#!/user/bin/perl
#List codecheck scripts
use File::Find;
use File::Copy;
use File::Temp;
use File::Basename;
use File::Spec;
use strict;

my $installDir = "C:/Program Files/SciTools";
my $validationScripts = "d:/temp/ScriptValidation/";
my $undExe = "$installDir/bin/pc-win64/und.exe";
my $uperlExe = "$installDir/bin/pc-win64/uperl.exe";
my $resultFile = "${validationScripts}verify_codecheck.csv";
my $limitFolder = "";  #Leave blank to test all checks, specify a folder to limit to a subset of checks

my @scriptBase;
push (@scriptBase,"$installDir/conf/plugin/User/Codecheck/");
push (@scriptBase,"$installDir/conf/plugin/SciTools/Codecheck/");


 my $tempFile = File::Temp->new();
 my $db = File::Temp->new();
 my $configFile = File::Temp->new();
 

 my $firstCheck=1;

 my $baseFolder;
 foreach my $scriptDir (@scriptBase){
  $baseFolder = $scriptDir;
  find(\&printCheckInfo,$scriptDir);
}
print "\nCompleted.\nResults are at $resultFile\n";

sub printCheckInfo{
  my $file = $File::Find::name;
  my $match = $baseFolder.$limitFolder;

  return if $limitFolder && $file !~ /$match/i;
      
  my $path =substr($File::Find::dir,length($baseFolder));
  return unless $file =~ /\.upl$/i;
  my $tempdir = File::Temp::tempdir( "CodeCheckValidatationXXXXXXX", CLEANUP => 1);
  
  my $base = basename($file);
  #Get the name of the test file
  my $testCode = $validationScripts.$path."/".$base;

  $testCode =~ s/\.upl$//i;
  if(-e $testCode.".cpp"){
    $testCode .= ".cpp";
  }elsif(-e $testCode.".C"){
    $testCode .= ".C";
  }elsif(-e $testCode.".c"){
    $testCode .= ".c";
  }elsif(-e $testCode.".h"){
    $testCode .= ".h";
  }elsif(-e $testCode.".adb"){
    $testCode .= ".adb";
  }elsif (-e $testCode){
    ;
  }else{
    return if $testCode =~ /Naming Conventions|Java/;
    print "Could not find test file for $testCode\n";
    return;
  }

 #Copy the check into a new file, we need to modify it and run it to get the name of the test 
 copy($file,$tempFile->filename);
 
  #Get formated name of test from the script.
  open (TEMP,">>$tempFile") || die ("couldn't open $tempFile $!\n");
  print TEMP "\n my \$pname = name();";
  print TEMP "\n my \$configName = \"$path/\$pname\";";
  print TEMP "\n \$configName =~ s/\\//\\\\/g;";
  print TEMP "\n \$configName =~ s/([^A-Za-z0-9-])/sprintf(\"%%%02X\", ord(\$1))/seg;";
  print TEMP "\n \$configName = \"single\\\\CheckState\\\\\".\$configName;";
  print TEMP "\n print \$configName;";
  close TEMP;
  printf ("Processing $file                      \r");
  my $testName = `"$uperlExe" "$tempFile"`;
  my $log = File::Temp->new();
  
  #Create Config File with formated name
  open (TEMP,">$configFile") || die ("couldn't open $configFile $!\n");
  print TEMP "[CodeCheckConfigNames]\n";
  print TEMP "$testName=2\n";
  close TEMP;
  
  #Create temp project and run codecheck on it
  my $cmd = "\"$undExe\" -db \"$db\" -quiet on create -languages C++ add \"$testCode\" analyze codecheck \"$configFile\" \"$tempdir\"";
  `$cmd  2>&1`;
  
  open (TEMP,"<$tempdir\\CodeCheckResultByTable.csv") || die ("\ncouldn't open results for  $path\\$base!\n");
  my @results = <TEMP>;
  close TEMP;
  
  if ($firstCheck){
    open (TEMP,">$resultFile") || die ("Could not write to master result file $resultFile. $!");
    foreach (@results){
        next unless $_ =~ /\S/;
        print TEMP $_;
    }
    close TEMP;
    $firstCheck=0;
  }else{
    open (TEMP,">>$resultFile") || die ("Could not write to master result file $resultFile. $!");
    my $firstline =1;
    foreach (@results){
        if ($firstline){
            $firstline=0;
            next;
        }
        next unless $_ =~ /\S/;
        print TEMP $_;
        
    }
    close TEMP;
  }
}
